***
*
*  Getsys.prg
*
*  Standard Clipper GET/READ Subsystem
*
*  Copyright (c) 1991-1993, Computer Associates International, Inc.
*  All rights reserved.
*
*  This version was modified for SOUND & VISION by Tom Groeger 1993
*
*     ReadKill( [<lKill>] )       --> lKill
*     ReadUpdated( [<lUpdated>] ) --> lUpdated
*     ReadFormat( [<bFormat>] )   --> bFormat | NIL
*
*  NOTE: compile with /m /n /w
*
*/

#include "Inkey.ch"
#include "Getexit.ch"
#include "SoundVis.ch"

#define K_UNDO          K_CTRL_U


//
// State variables for active READ
//
STATIC sbFormat
STATIC slUpdated := .F.
STATIC slKillRead
STATIC slBumpTop
STATIC slBumpBot
STATIC snLastExitState
STATIC snLastPos
STATIC soActiveGet
STATIC scReadProcName
STATIC snReadProcLine

//@@New:
static nPos
static nBouncePos
static nFieldPos
static lMouseStat
static lListIcon


//
// Format of array used to preserve state variables
//
#define GSV_KILLREAD       1
#define GSV_BUMPTOP        2
#define GSV_BUMPBOT        3
#define GSV_LASTEXIT       4
#define GSV_LASTPOS        5
#define GSV_ACTIVEGET      6
#define GSV_READVAR        7
#define GSV_READPROCNAME   8
#define GSV_READPROCLINE   9

//@@New :
#define GSV_NPOS        10              // Aktiv GetField
#define GSV_BOUNCEPOS   11              // New Getfield
#define GSV_FIELDPOS    12              // Position within Get
#define GSV_MOUSESTAT   13              // Hide Mouse .t./.f.
#define GSV_LISTICON    14              // ListIcon active .t./.f.
#define GSV_COUNT       14              // Size of SaveArrays


#define GE_BOUNCE        9              // ExitState: aktivate new Get


/***
*
*  ReadModal()
*
*  Standard modal READ on an array of GETs
*
*/
FUNCTION ReadModal( GetList, nPos )

   LOCAL oGet
   LOCAL aSavGetSysVars

   IF ( VALTYPE( sbFormat ) == "B" )
      EVAL( sbFormat )
   ENDIF

   IF ( EMPTY( GetList ) )
      
      // S'87 compatibility
      SETPOS( MAXROW() - 1, 0 )
      RETURN (.F.)                  // NOTE

   ENDIF

   // Preserve state variables
   aSavGetSysVars := ClearGetSysVars()

   // Set these for use in SET KEYs
   scReadProcName := PROCNAME( 1 )
   snReadProcLine := PROCLINE( 1 )

   // Set initial GET to be read
   IF !( VALTYPE( nPos ) == "N" .AND. nPos > 0 )
      nPos := Settle( Getlist, 0 )
   ENDIF

   WHILE !( nPos == 0 )

      // Get next GET from list and post it as the active GET
      PostActiveGet( oGet := GetList[ nPos ] )

      // Read the GET  @@new: Parameter Getlist
      IF ( VALTYPE( oGet:reader ) == "B" )
         EVAL( oGet:reader, oGet, getlist )    // Use custom reader block
      ELSE
         GetReader( oGet, getlist )  // Use standard reader
      ENDIF

      // Move to next GET based on exit condition
      nPos := Settle( GetList, nPos )

   ENDDO


   // Restore state variables
   RestoreGetSysVars( aSavGetSysVars )

   // S'87 compatibility
   SETPOS( MAXROW() - 1, 0 )

   RETURN ( slUpdated )



/***
*
*  GetReader()
*
*  Standard modal read of a single GET
*  @@New: Parameter getlist
*/
PROCEDURE GetReader( oGet, Getlist )

   // Read the GET if the WHEN condition is satisfied
   IF ( GetPreValidate( oGet ) )

      //@@New : Activate the GET for reading
      GetSetFocus ( @oGet )

      WHILE ( oGet:exitState == GE_NOEXIT )

         // Check for initial typeout (no editable positions)
         IF ( oGet:typeOut )
            oGet:exitState := GE_ENTER
         ENDIF

         // @@new: Switch Cursor ON
         SETCURSOR (1)

         // Apply keystrokes until exit
         WHILE ( oGet:exitState == GE_NOEXIT )

            // @@New switch Mouse on, Parameter getlist
               Mse_Show ( .t. )
               GetApplyKey( oGet, Inkey(0), getlist  )

         ENDDO

         SETCURSOR (0)


         // Disallow exit if the VALID condition is not satisfied
         IF ( !GetPostValidate( oGet ) )
            oGet:exitState := GE_NOEXIT
         ENDIF
      ENDDO

      // De-activate the GET
      oGet:killFocus()

   ENDIF

   RETURN



/***
*
*  GetApplyKey()
*
*  Apply a single INKEY() keystroke to a GET
*
*  NOTE: GET must have focus.
*  @@NEW: Parameter getlist
*/
PROCEDURE GetApplyKey( oGet, nKey, getlist )

   LOCAL cKey
   LOCAL bKeyBlock

// @@New  Hide Mouse
// ------------------
   Mse_Show (.f.)

// Disable MouseRequest
// --------------------
   IF nkey != K_MOUSE


      // Check for SET KEY first
      IF !( ( bKeyBlock := setkey( nKey ) ) == NIL )
         GetDoSetKey( bKeyBlock, oGet )
         RETURN                           // NOTE
      ENDIF

   ENDIF

   DO CASE

   //@New
   CASE ( nkey == K_MOUSE )
        Get_Mse ( Getlist )

   CASE ( nKey == K_UP )
      oGet:exitState := GE_UP

   CASE ( nKey == K_SH_TAB )
      oGet:exitState := GE_UP

   CASE ( nKey == K_DOWN )
      oGet:exitState := GE_DOWN

   CASE ( nKey == K_TAB )
      oGet:exitState := GE_DOWN

   CASE ( nKey == K_ENTER )
      oGet:exitState := GE_ENTER

   CASE ( nKey == K_ESC )
      IF ( SET( _SET_ESCAPE ) )
         
         oGet:undo()
         oGet:exitState := GE_ESCAPE

      ENDIF

   CASE ( nKey == K_PGUP )
      oGet:exitState := GE_WRITE

   CASE ( nKey == K_PGDN )
      oGet:exitState := GE_WRITE

   CASE ( nKey == K_CTRL_HOME )
      oGet:exitState := GE_TOP


#ifdef CTRL_END_SPECIAL

   // Both ^W and ^End go to the last GET
   CASE ( nKey == K_CTRL_END )
      oGet:exitState := GE_BOTTOM

#else

   // Both ^W and ^End terminate the READ (the default)
   CASE ( nKey == K_CTRL_W )
      oGet:exitState := GE_WRITE

#endif


   CASE ( nKey == K_INS )
      SET( _SET_INSERT, !SET( _SET_INSERT ) )
      ShowScoreboard()

   CASE ( nKey == K_UNDO )
      oGet:undo()

   CASE ( nKey == K_HOME )
      oGet:home()

   CASE ( nKey == K_END )
      oGet:end()

   CASE ( nKey == K_RIGHT )
      oGet:right()

   CASE ( nKey == K_LEFT )
      oGet:left()

   CASE ( nKey == K_CTRL_RIGHT )
      oGet:wordRight()

   CASE ( nKey == K_CTRL_LEFT )
      oGet:wordLeft()

   CASE ( nKey == K_BS )
      oGet:backSpace()

   CASE ( nKey == K_DEL )
      oGet:delete()

   CASE ( nKey == K_CTRL_T )
      oGet:delWordRight()

   CASE ( nKey == K_CTRL_Y )
      oGet:delEnd()

   CASE ( nKey == K_CTRL_BS )
      oGet:delWordLeft()

   OTHERWISE

      IF ( nKey >= 32 .AND. nKey <= 255 )

         cKey := CHR( nKey )

         IF ( oGet:type == "N" .AND. ( cKey == "." .OR. cKey == "," ) )
            oGet:toDecPos()
         ELSE
            
            IF ( SET( _SET_INSERT ) )
               oGet:insert( cKey )
            ELSE
               oGet:overstrike( cKey )
            ENDIF

            IF ( oGet:typeOut )
               IF ( SET( _SET_BELL ) )
                  ?? CHR(7)
               ENDIF

               IF ( !SET( _SET_CONFIRM ) )
                  oGet:exitState := GE_ENTER
               ENDIF
            ENDIF

         ENDIF

      ENDIF

   ENDCASE

   RETURN



/***
*
*  GetPreValidate()
*
*  Test entry condition (WHEN clause) for a GET
*
*/
FUNCTION GetPreValidate( oGet )

   LOCAL lSavUpdated
   LOCAL lWhen := .T.

   IF !( oGet:preBlock == NIL )

      lSavUpdated := slUpdated

      lWhen := EVAL( oGet:preBlock, oGet )

      oGet:display()

      ShowScoreBoard()
      slUpdated := lSavUpdated

   ENDIF

   IF ( slKillRead )
      
      lWhen := .F.
      oGet:exitState := GE_ESCAPE       // Provokes ReadModal() exit

   ELSEIF ( !lWhen )
      
      oGet:exitState := GE_WHEN         // Indicates failure

   ELSE
      
      oGet:exitState := GE_NOEXIT       // Prepares for editing

   END

   RETURN ( lWhen )



/***
*
*  GetPostValidate()
*
*  Test exit condition (VALID clause) for a GET
*
*  NOTE: Bad dates are rejected in such a way as to preserve edit buffer
*
*/
FUNCTION GetPostValidate( oGet )

   LOCAL lSavUpdated
   LOCAL lValid := .T.


   IF ( oGet:exitState == GE_ESCAPE )
      RETURN ( .T. )                   // NOTE
   ENDIF

   IF ( oGet:badDate() )
      oGet:home()
      DateMsg()
      ShowScoreboard()
      RETURN ( .F. )                   // NOTE
   ENDIF

   // If editing occurred, assign the new value to the variable
   IF ( oGet:changed )
      oGet:assign()
      slUpdated := .T.
   ENDIF

   // Reform edit buffer, set cursor to home position, redisplay
   oGet:reset()

   // Check VALID condition if specified
   IF !( oGet:postBlock == NIL )

      lSavUpdated := slUpdated

      // S'87 compatibility
      SETPOS( oGet:row, oGet:col + LEN( oGet:buffer ) )

      lValid := EVAL( oGet:postBlock, oGet )

      // Reset S'87 compatibility cursor position
      SETPOS( oGet:row, oGet:col )

      ShowScoreBoard()
      oGet:updateBuffer()

      slUpdated := lSavUpdated

      IF ( slKillRead )
         oGet:exitState := GE_ESCAPE      // Provokes ReadModal() exit
         lValid := .T.

      ENDIF
   ENDIF

   RETURN ( lValid )



/***
*
*  GetDoSetKey()
*
*  Process SET KEY during editing
*
*/
PROCEDURE GetDoSetKey( keyBlock, oGet )

   LOCAL lSavUpdated

   // If editing has occurred, assign variable
   IF ( oGet:changed )
      oGet:assign()
      slUpdated := .T.
   ENDIF

   lSavUpdated := slUpdated

   EVAL( keyBlock, scReadProcName, snReadProcLine, ReadVar() )

   ShowScoreboard()
   oGet:updateBuffer()

   slUpdated := lSavUpdated

   IF ( slKillRead )
      oGet:exitState := GE_ESCAPE      // provokes ReadModal() exit
   ENDIF

   RETURN





/***
*              READ services
*/



/***
*
*  Settle()
*
*  Returns new position in array of Get objects, based on:
*     - current position
*     - exitState of Get object at current position
*
*  NOTES: return value of 0 indicates termination of READ
*         exitState of old Get is transferred to new Get
*
*/
STATIC FUNCTION Settle( GetList, nPos )

   LOCAL nExitState

   IF ( nPos == 0 )
      nExitState := GE_DOWN
   ELSE
      nExitState := GetList[ nPos ]:exitState
   ENDIF

   IF ( nExitState == GE_ESCAPE .or. nExitState == GE_WRITE )
      RETURN ( 0 )               // NOTE
   ENDIF

   IF !( nExitState == GE_WHEN )
      // Reset state info
      snLastPos := nPos
      slBumpTop := .F.
      slBumpBot := .F.
   ELSE
      // Re-use last exitState, do not disturb state info
      nExitState := snLastExitState
   ENDIF

   //
   // Move
   //
   DO CASE
   CASE ( nExitState == GE_UP )
      nPos--

   CASE ( nExitState == GE_DOWN )
      nPos++

   CASE ( nExitState == GE_TOP )
      nPos       := 1
      slBumpTop  := .T.
      nExitState := GE_DOWN

   CASE ( nExitState == GE_BOTTOM )
      nPos       := LEN( GetList )
      slBumpBot  := .T.
      nExitState := GE_UP

   CASE ( nExitState == GE_ENTER )
      nPos++

   // @@New ( Mouse )
   case ( nExitState == GE_BOUNCE )
      nPos := nBouncePos
      nExitState := GE_DOWN


   ENDCASE

   //
   // Bounce
   //
   IF ( nPos == 0 )                       // Bumped top
      IF ( !ReadExit() .and. !slBumpBot )
         slBumpTop  := .T.
         nPos       := snLastPos
         nExitState := GE_DOWN
      ENDIF

   ELSEIF ( nPos == len( GetList ) + 1 )  // Bumped bottom
      IF ( !ReadExit() .and. !( nExitState == GE_ENTER ) .and. !slBumpTop )
         slBumpBot  := .T.
         nPos       := snLastPos
         nExitState := GE_UP
      ELSE
         nPos := 0
      ENDIF
   ENDIF

   // Record exit state
   snLastExitState := nExitState

   IF !( nPos == 0 )
      GetList[ nPos ]:exitState := nExitState
   ENDIF
   
   RETURN ( nPos )



/***
*
*  PostActiveGet()
*
*  Post active GET for ReadVar(), GetActive()
*
*/
STATIC PROCEDURE PostActiveGet( oGet )

   GetActive( oGet )
   ReadVar( GetReadVar( oGet ) )

   ShowScoreBoard()

   RETURN



/***
*
*  ClearGetSysVars()
*
*  Save and clear READ state variables. Return array of saved values
*
*  NOTE: 'Updated' status is cleared but not saved (S'87 compatibility)
*/
STATIC FUNCTION ClearGetSysVars()

   LOCAL aSavSysVars[ GSV_COUNT ]

   // Save current sys vars
   aSavSysVars[ GSV_KILLREAD ]     := slKillRead
   aSavSysVars[ GSV_BUMPTOP ]      := slBumpTop
   aSavSysVars[ GSV_BUMPBOT ]      := slBumpBot
   aSavSysVars[ GSV_LASTEXIT ]     := snLastExitState
   aSavSysVars[ GSV_LASTPOS ]      := snLastPos
   aSavSysVars[ GSV_ACTIVEGET ]    := GetActive( NIL )
   aSavSysVars[ GSV_READVAR ]      := ReadVar( "" )
   aSavSysVars[ GSV_READPROCNAME ] := scReadProcName
   aSavSysVars[ GSV_READPROCLINE ] := snReadProcLine

// @@New
// -----
   aSavSysVars [GSV_NPOS]           := nPos
   aSavSysVars [GSV_BOUNCEPOS]      := nBouncePos
   aSavSysVars [GSV_FIELDPOS]       := nFieldPos
   aSavSysVars [GSV_MOUSESTAT]      := MSE_Show (.f.)
   aSavSysVars [GSV_LISTICON]       := lListIcon


// Re-init old ones
   slKillRead      := .F.
   slBumpTop       := .F.
   slBumpBot       := .F.
   snLastExitState := 0
   snLastPos       := 0
   scReadProcName  := ""
   snReadProcLine  := 0
   slUpdated       := .F.

//@New
   nPos            := 0
   nBouncePos      := 0
   nFieldPos       := -1
   lMouseStat      := .f.
   lListIcon       := .f.

   RETURN ( aSavSysVars )



/***
*
*  RestoreGetSysVars()
*
*  Restore READ state variables from array of saved values
*
*  NOTE: 'Updated' status is not restored (S'87 compatibility)
*
*/
STATIC PROCEDURE RestoreGetSysVars( aSavSysVars )

   slKillRead      := aSavSysVars[ GSV_KILLREAD ]
   slBumpTop       := aSavSysVars[ GSV_BUMPTOP ]
   slBumpBot       := aSavSysVars[ GSV_BUMPBOT ]
   snLastExitState := aSavSysVars[ GSV_LASTEXIT ]
   snLastPos       := aSavSysVars[ GSV_LASTPOS ]

   GetActive( aSavSysVars[ GSV_ACTIVEGET ] )

   ReadVar( aSavSysVars[ GSV_READVAR ] )

   scReadProcName  := aSavSysVars[ GSV_READPROCNAME ]
   snReadProcLine  := aSavSysVars[ GSV_READPROCLINE ]

// @@New:
// -----
   nPos         := aSavSysVars [GSV_NPOS]
   nBouncePos   := aSavSysVars [GSV_BOUNCEPOS]
   nFieldPos    := aSavSysVars [GSV_FIELDPOS]
   lMouseStat   := aSavSysVar  [GSV_MOUSESTAT]
   lListIcon    := aSavSysVar  [GSV_LISTICON]

   MSE_Show ( lMouseStat )


   RETURN



/***
*
*  GetReadVar()
*
*  Set READVAR() value from a GET
*
*/
STATIC FUNCTION GetReadVar( oGet )

   LOCAL cName := UPPER( oGet:name )
   LOCAL i

   // The following code includes subscripts in the name returned by
   // this FUNCTIONtion, if the get variable is an array element
   //
   // Subscripts are retrieved from the oGet:subscript instance variable
   //
   // NOTE: Incompatible with Summer 87
   //
   IF !( oGet:subscript == NIL )
      FOR i := 1 TO LEN( oGet:subscript )
         cName += "[" + LTRIM( STR( oGet:subscript[i] ) ) + "]"
      NEXT
   END

   RETURN ( cName )





/***
*              System Services
*/



/***
*
*  __SetFormat()
*  
*  SET FORMAT service
*
*/
PROCEDURE __SetFormat( b )
   sbFormat := IF( VALTYPE( b ) == "B", b, NIL )
   RETURN



/***
*
*  __KillRead()
*
*  CLEAR GETS service
*
*/
PROCEDURE __KillRead()
   slKillRead := .T.
   RETURN



/***
*
*  GetActive()
*
*  Retrieves currently active GET object
*/
FUNCTION GetActive( g )

   LOCAL oldActive := soActiveGet

   IF ( PCOUNT() > 0 )
      soActiveGet := g
   ENDIF

   RETURN ( oldActive )



/***
*
*  Updated()
*
*/
FUNCTION Updated()
   RETURN slUpdated



/***
*
*  ReadExit()
*
*/
FUNCTION ReadExit( lNew )
   RETURN ( SET( _SET_EXIT, lNew ) )



/***
*
*  ReadInsert()
*
*/
FUNCTION ReadInsert( lNew )
   RETURN ( SET( _SET_INSERT, lNew ) )



/***
*              Wacky Compatibility Services
*/


// Display coordinates for SCOREBOARD
#define SCORE_ROW      0
#define SCORE_COL      60


/***
*
*  ShowScoreboard()
*
*/
STATIC PROCEDURE ShowScoreboard()

   LOCAL nRow
   LOCAL nCol

   IF ( SET( _SET_SCOREBOARD ) )
      nRow := ROW()
      nCol := COL()

      SETPOS( SCORE_ROW, SCORE_COL )
      DISPOUT( IF( SET( _SET_INSERT ), "Ins", "   " ) )
      SETPOS( nRow, nCol )
   ENDIF

   RETURN



/***
*
*  DateMsg()
*
*/
STATIC PROCEDURE DateMsg()

   LOCAL nRow
   LOCAL nCol

   IF ( SET( _SET_SCOREBOARD ) )
      
      nRow := ROW()
      nCol := COL()

      SETPOS( SCORE_ROW, SCORE_COL )
      DISPOUT( "Invalid Date" )
      SETPOS( nRow, nCol )

      WHILE ( NEXTKEY() == 0 )
      END

      SETPOS( SCORE_ROW, SCORE_COL )
      DISPOUT( SPACE( 12 ) )
      SETPOS( nRow, nCol )

   ENDIF

   RETURN



/***
*
*  RangeCheck()
*
*  NOTE: Unused second param for 5.00 compatibility.
*
*/
FUNCTION RangeCheck( oGet, junk, lo, hi )

   LOCAL cMsg, nRow, nCol
   LOCAL xValue

   IF ( !oGet:changed )
      RETURN ( .T. )          // NOTE
   ENDIF

   xValue := oGet:varGet()

   IF ( xValue >= lo .and. xValue <= hi )
      RETURN ( .T. )          // NOTE
   ENDIF

   IF ( SET(_SET_SCOREBOARD) )
      
      cMsg := "Range: " + LTRIM( TRANSFORM( lo, "" ) ) + ;
            " - " + LTRIM( TRANSFORM( hi, "" ) )

      IF ( LEN( cMsg ) > MAXCOL() )
         cMsg := SUBSTR( cMsg, 1, MAXCOL() )
      ENDIF

      nRow := ROW()
      nCol := COL()

      SETPOS( SCORE_ROW, MIN( 60, MAXCOL() - LEN( cMsg ) ) )
      DISPOUT( cMsg )
      SETPOS( nRow, nCol )

      WHILE ( NEXTKEY() == 0 )
      END

      SETPOS( SCORE_ROW, MIN( 60, MAXCOL() - LEN( cMsg ) ) )
      DISPOUT( SPACE( LEN( cMsg ) ) )
      SETPOS( nRow, nCol )

   ENDIF

   RETURN ( .F. )



/***
*
*  ReadKill()
*
*/
FUNCTION ReadKill( lKill )

   LOCAL lSavKill := slKillRead

   IF ( PCOUNT() > 0 )
      slKillRead := lKill
   ENDIF

   RETURN ( lSavKill )



/***
*
*  ReadUpdated()
*
*/
FUNCTION ReadUpdated( lUpdated )
   
   LOCAL lSavUpdated := slUpdated
   
   IF ( PCOUNT() > 0 )
      slUpdated := lUpdated
   ENDIF

   RETURN ( lSavUpdated )
      


/***
*
*  ReadFormat()
*
*/
FUNCTION ReadFormat( b )
   
   LOCAL bSavFormat := sbFormat

   IF ( PCOUNT() > 0 )
      sbFormat := b
   ENDIF

   RETURN ( bSavFormat )
      


/*

  SUBROUTINE GetSetFocus   /@@NEW                     
  This Function sets the Input-Focus to the GET       
  and positions the Cursor at the MouseCol            
                                                      
  PARAMETER: oGet :  active Get                       
  RETURNS :  nil                                      

*/

FUNCTION GetSetFocus( oGet )

   // donate Input-Focus
   // -------------------
   oGet:SetFocus()

   // fix FieldPosition
   // -----------------
   IF ( nFieldPos != -1 )

      IF ( LEN( TRANSFORM ( oGet:varGet, oGet:picture) ) > nFieldPos )
         oGet:pos := nFieldPos     // New Fieldposition
         oGet:Display ( )           // display it

      ENDIF
   ENDIF

   nFieldPos := -1

RETURN ( nil )



/*

  SUBROUTINE Get_Mse      /@@NEW                      
  Function selects the GET by MouseCursor             
                                                      
                                                      
  PARAMETER: Getlist                                  
  RETURNS:   nil                                      

*/
FUNCTION Get_Mse ( getlist )

  LOCAL nMouseCol := Mse_Col()

  LOCAL nMouseKey := Mse_Key()

  LOCAL nGetFld, nGetLen, nX, oGet


    DO CASE
    CASE ( nMouseKey == 1)

        // Left MouseButton
        // which Get/Button would you like ?
        // ---------------------------------
         FOR nGetFld := LEN( getlist ) TO 1 STEP -1

             oGet := getlist [ nGetFld ]
             nGetLen := LEN ( TRANSFORM ( oGet:varGet, oGet:picture ))

             // Buttons ?
             IF ( nX := oMouseButt ( oGet:cargo )) # 0
                oGet:varput ( nX )
                EXIT

             // or GetFields ?
             ELSEIF Mse_Check ( oGet:row, oGet:col,, oGet:col+nGetLen )
                EXIT

             ENDIF

         NEXT

         DO CASE

           // Mouse hits Getfield
           // -------------------
         CASE nGetFld != 0

           // select new Get
              nBouncePos := nGetFld
              soActiveGet:exitState := GE_BOUNCE
      
           // Get Position within GetField
              nFieldPos := nMouseCol - oGet:col + 1


           // Mouse hits WindowIcon ?
           // ----------------------
         CASE ( nX := oMouseWin ()) # 0

              IF nX == M_CLOSE
                 soActiveGet:exitState := GE_ESCAPE
              ENDIF


           // ListIcon active ?
           // -----------------
           CASE lListIcon .AND. Mse_Check ( oGet:row, ;
                                            oGet:col+nGetLen+2,,;
                                            oGet:col+nGetLen+3 )
                oGet:exitState := GE_ENTER

           // Missed again
         OTHERWISE
              TONE ( 1000 )

         ENDCASE


    CASE ( nMouseKey == 2 )
         // Right Button = ESCAPE
         soActiveGet:exitState := GE_ESCAPE


    CASE ( nMouseKey == 4 )
         // Center Button = ENTER
         soActiveGet:exitState := GE_WRITE

  ENDCASE

RETURN ( nil )




/*

  SUBROUTINE ReadaButton  /@@NEW                      
  Reader for GET-oButtons                             
                                                      
  PARAMETER: oGet     selected  Get-Objekt            
             nSec     TimeOut-Value                   
             getlist  You know this                   
  RETURNS :  nil                                      

*/

FUNCTION ReadaButton ( oGet , nSec, getlist )

   LOCAL nChoice := oGet:VarGet, nLen := oCountButt ( oGet:cargo )

   LOCAL nX, nKey, bKeyBlock, xBlockRet

   LOCAL nCurs := SETCURSOR (0)


   /* TimeOut set ? */
   DEFAULT nSec TO 0

   // How about the WHEN-Clause ?
   // ---------------------------
   IF ( GetPreValidate(oGet) )

      // Activate she says
      // -----------------
      oGet:SetFocus()
      oDisButt ( oGet:cargo, MAX( 1, nChoice) )

      DO WHILE ( oGet:exitState == GE_NOEXIT )

         MSE_Show ( .t. ) 
         nKey := INKEY(nSec)
         MSE_Show ( .f. )

         IF nKey == 0
            nChoice := B_SKIP
            oGet:exitState := GE_ESCAPE
         ENDIF

      // Scan KeyCode
         DO CASE 

         // Is there any SetKey ? 
         CASE ( (bKeyBlock := SetKey(nkey)) <> NIL )
              xBlockRet := EVAL ( bKeyBlock, oGet )

              // We leave the show, if Function returns a .TRUE.
              IF VALTYPE(xBlockRet) == 'L'
                 IF xBlockRet
                    nChoice := B_SKIP
                    oGet:ExitState := GE_ESCAPE
                 ENDIF
              ENDIF

         CASE nKey == K_MOUSE

              DO CASE
              // Mouse over Button ?
              CASE ( nX := oMouseButt( oGet:cargo)) # 0
                   nChoice := oSetFocus ( oGet:cargo, nX )
                   nKey := K_ENTER

              // Mouse over Icon ?
              CASE ( nX := oMouseWin ()) # 0
                   nChoice := nX
                   oGet:exitState := GE_ESCAPE

              // Mouse over Gets ?
              Otherwise
                   Get_Mse ( getlist )

              ENDCASE

              // HotKey ?
         CASE (nX := oKeyButt ( oGet:cargo, nKey )) # 0
              nChoice := oSetFocus ( oGet:cargo, nX )
              nKey := K_ENTER


         CASE nKey == K_UP .OR. nKey == K_SH_TAB
              nX := oSetFocus ( oGet:cargo, nChoice-1 )

              IF nX == 0
                 oGet:exitstate := GE_UP

              ELSE
                 nChoice := nX

              ENDIF

         CASE nKey == K_DOWN .OR. nKey == K_TAB
              nX := oSetFocus ( oGet:cargo, nChoice+1 )

              IF nX > nLen
                 oGet:exitstate := GE_DOWN

              ELSE
                 nChoice := nX

              ENDIF


         CASE nKey == K_LEFT .OR. nKey == K_RIGHT
              nChoice := oSetFocus ( oGet:cargo,;
                         WrapWert ( nKey == K_RIGHT, nChoice, 1, nLen) )


         CASE nKey == K_HOME .OR. nKey == K_END
              nChoice := oSetFocus ( oGet:cargo,;
                     IIF (nKey == K_HOME,1, nLen) )

        
         CASE nKey == K_ESC
              nChoice := B_ESC
              oGet:exitState := GE_ESCAPE

         ENDCASE

         // Shall we push the Button ?
         IF nKey == K_ENTER
             nChoice := oPushButt ( oGet:cargo )
             oGet:exitState := GE_ENTER

         ENDIF

      ENDDO


      // store GetVar
      // --------------
      oGet:VarPut( nChoice )

      // steal Focus
      // -----------
      oGet:KillFocus()
      oDisButt ( oGet:cargo )

      SETCURSOR ( nCurs )

   ENDIF

 RETURN ( nil )

/*

  SUBROUTINE FrameReader  /@@NEW                      
  Reader for framed GET's / Listboxes                 
                                                      
  PARAMETER: oGet    active Get-Objekt                
             nColor  Select-Color (0-255)             
             n2Color Unselect-Color (0-255)           
             aList   Listbox Array                    
             getlist we need it for getapplykey       
                                                      
  RETURNS :  nil                                      

*/

FUNCTION FrameReader (oGet, nColor, n2Color, aList, getlist )
   LOCAL nTop, nLeft, nBottom, nRight
   LOCAL cVar  := oGet:varget()
   LOCAL nGetLen, nCurs, nX


   // read GET only WHEN
   IF ( GetPreValidate(oGet) )

      // set SysVar ListIcon
      lListIcon := ( aList # nil )

      nGetLen := LEN( TRANSFORM (oGet:varGet, oGet:picture))

      // Frame Roger Rabbit
      IF nColor # nil
         OutLine ( oGet:row, oGet:col-1, SPACE(nGetLen), 2, nColor )
      ENDIF


      // activate GET
      GetSetFocus ( @oGet )

      IF lListIcon

         // show Icon
         ShowStr ( oGet:row, oGet:col+nGetLen+2, A_ICONBOX, 7 )

         // set Positions
         nTop    := oGet:row + 2
         nLeft   := oGet:col + 1
         nRight  := oGet:col + nGetLen + 1

         // adjust lower Border
         nBottom := MIN( MaxRow()-2, nTop+LEN( aList )-1)
         IF nBottom - nTop < 3
            nTop := nBottom - 3
         ENDIF


      ENDIF


      // let's have some Input
      DO WHILE ( oGet:exitState == GE_NOEXIT )

         // check TypeOut
         IF ( oGet:typeOut )
            oGet:exitState := GE_ENTER
         ENDIF

         // we need a Cursor
         nCurs := SETCURSOR (1)

         // get KeyCode
         DO WHILE ( oGet:exitState == GE_NOEXIT )

            Mse_Show ( .t. )
            GetApplyKey( oGet, INKEY(0), getlist )

         ENDDO

         SETCURSOR ( nCurs )

         IF ( oGet:exitstate == GE_ENTER .AND. lListIcon )
            nX := ASCAN ( aList, {|c| c == TRIM( cVar )})

            // Activate Listbox
            cCurs := SETCURSOR ( 0 )
            nX := ;
            ChoseBox ( nTop, nLeft, nBottom, nRight+1, aList,IIF( nX==0,1,nX ))
            SETCURSOR ( cCurs )

            // Check RetValue
            IF nX > 0
               cVar := aList [nX]
               oGet:VarPut ( PADR(cVar, nGetLen))
               oGet:updateBuffer()

            ELSE
               oGet:exitState := GE_NOEXIT

            ENDIF
         ENDIF

         // Don't leave me now ( if the VALID-Option is .false. )
         IF ( !GetPostValidate(oGet) )
            oGet:exitState := GE_NOEXIT
         ENDIF

      ENDDO

      // kill Icon and Outline
      IF lListIcon
         @ oGet:row, oGet:col+nGetLen+2 SAY '  '
      ENDIF

      IF nColor # nil
         DelLine ( oGet:row, oGet:col-1, nGetLen, n2color, .f.,lListIcon )
      ENDIF

      // reset ListIconFlag
      lListIcon := .f.

      // steal Focus
      oGet:KillFocus()


   ENDIF

RETURN (nil)

